using System;
using AVFoundation;
using static Barcode2DScanner;
using CoreGraphics;
using Foundation;
using UIKit;
using CoreMedia;
using CoreImage;
using CoreVideo;
using System.Runtime.InteropServices;



namespace parserTest
{

    enum eMainScreenState
    {
        NORMAL,
        LAUNCHING_CAMERA,
        CAMERA,
        CAMERA_DECODING,
        DECODE_DISPLAY,
        CANCELLING
    };


    public class DecoderResult: NSObject {
        bool success;
        NSString result;

        public static DecoderResult createSuccess(NSString result) {
            DecoderResult obj = new DecoderResult();

                obj.success = true;
                obj.result = result;

            return obj;
        }

        public static DecoderResult createFailure(NSString result) {
            DecoderResult obj = new DecoderResult();
            obj.success = false;
            obj.result = result;
            return obj;
        }
    }

    public partial class ScannerViewController : UIViewController
    {
        partial void torchBtn(UIButton sender)
        {
            toggleTorch();
        }

        partial void close(UIButton sender)
        {
            this.stopRunning();
            this.DismissViewController(true, null);
        }

        AVCaptureSession captureSession;
        AVCaptureDevice videoDevice;
        AVCaptureDeviceInput videoInput;
        AVCaptureVideoPreviewLayer previewLayer;
        bool running;
        NSTimer focusTimer;
        bool detected;
        NSMutableArray allowedBarcodeTypes;
        eMainScreenState state;

        public static NSString DecoderResultNotification = new NSString("DecoderResultNotification");

        public ScannerViewController(IntPtr handle) : base(handle)
        {
        }

        public override void ViewDidLoad()
        {
            base.ViewDidLoad();
            setupCaptureSession();
            previewLayer.Frame = this.View.Layer.Bounds;
            this.View.Layer.AddSublayer(previewLayer);

            CustomeOverlay();
            focusTimer = NSTimer.CreateRepeatingScheduledTimer(0.7, delegate { reFocus(); });

            this.allowedBarcodeTypes = new NSMutableArray();
            this.allowedBarcodeTypes.Add(new NSString("org.iso.QRCode"));
            this.allowedBarcodeTypes.Add(new NSString("org.iso.PDF417"));
            this.allowedBarcodeTypes.Add(new NSString("org.gs1.UPC-E"));
            this.allowedBarcodeTypes.Add(new NSString("org.iso.Aztec"));
            this.allowedBarcodeTypes.Add(new NSString("org.iso.Code39"));
            this.allowedBarcodeTypes.Add(new NSString("org.iso.Code39Mod43"));
            this.allowedBarcodeTypes.Add(new NSString("org.gs1.EAN-13"));
            this.allowedBarcodeTypes.Add(new NSString("org.gs1.EAN-8"));
            this.allowedBarcodeTypes.Add(new NSString("com.intermec.Code93"));
            this.allowedBarcodeTypes.Add(new NSString("org.iso.Code128"));
        }

        public override void DidReceiveMemoryWarning()
        {
            base.DidReceiveMemoryWarning();
            // Release any cached data, images, etc that aren't in use.
        }

        public override void ViewDidAppear(bool animated)
        {
            base.ViewDidAppear(animated);
            startRunning();
        }

        public void viewWillDisappear(bool animated)
        {
            base.ViewWillDisappear(animated);
            stopRunning();
            if (focusTimer != null)
            {
                focusTimer.Invalidate();
                focusTimer = null;
            }
        }

        void CustomeOverlay()
        {
            this.View.BringSubviewToFront(closeBtn);
            this.View.BringSubviewToFront(demoLbl);
            this.View.BringSubviewToFront(torch);
        }

        void setupCaptureSession()
        {
            if (captureSession != null)
                return;
            videoDevice = AVCaptureDevice.GetDefaultDevice(AVMediaType.Video); 
            if (videoDevice == null)
            {
                Console.WriteLine("No video camera on this device!");
                return;
            }
            captureSession = new AVCaptureSession();
            videoInput = AVCaptureDeviceInput.FromDevice(videoDevice);
            if (captureSession.CanAddInput(videoInput))
            {
                captureSession.AddInput(videoInput);
            }
            previewLayer = new AVCaptureVideoPreviewLayer(captureSession);
            previewLayer.VideoGravity = AVLayerVideoGravity.ResizeAspectFill;

            AVCaptureVideoDataOutput captureOutput = new AVCaptureVideoDataOutput();
            captureOutput.AlwaysDiscardsLateVideoFrames = true;
            CaptureOutputDelegate outputRecorder = new CaptureOutputDelegate(this);
            captureOutput.SetSampleBufferDelegateQueue(outputRecorder, CoreFoundation.DispatchQueue.MainQueue);

            if (captureSession.CanAddOutput(captureOutput))
            {
                captureSession.AddOutput(captureOutput);
            }
        }

        void startRunning()
        {
            if (running)
                return;
            this.state = eMainScreenState.LAUNCHING_CAMERA;
            this.captureSession.StartRunning();
            this.previewLayer.Hidden = false;
            this.running = true;
            this.state = eMainScreenState.CAMERA;
        }

        void stopRunning()
        {
            if (!running)
                return;
            captureSession.StopRunning();
            running = false;
            this.state = eMainScreenState.NORMAL;
        }

        void reFocus()
        {
            NSError error;
            if (videoDevice.LockForConfiguration(out error))
            {
                if (videoDevice.FocusPointOfInterestSupported)
                {
                    videoDevice.FocusPointOfInterest = new CGPoint(0.49, 0.49);
                    videoDevice.FocusMode = AVCaptureFocusMode.AutoFocus;
                }
                videoDevice.UnlockForConfiguration();
            }
        }

        void toggleTorch()
        {
            if (videoDevice.IsTorchModeSupported(AVCaptureTorchMode.On))
            {
                NSError error;

                if (videoDevice.LockForConfiguration(out error))
                {
                    if (videoDevice.TorchMode == AVCaptureTorchMode.On)
                        videoDevice.TorchMode = AVCaptureTorchMode.Off;
                    else
                        videoDevice.TorchMode = AVCaptureTorchMode.On;

                    if (videoDevice.IsFocusModeSupported(AVCaptureFocusMode.ContinuousAutoFocus))
                        videoDevice.FocusMode = AVCaptureFocusMode.ContinuousAutoFocus;

                    videoDevice.UnlockForConfiguration();
                }
                else
                {

                }
            }
        }

        public class CaptureOutputDelegate : AVCaptureVideoDataOutputSampleBufferDelegate
        {

            private ScannerViewController parent { get; set; }
            int width;
            int height;
            private UIImageView imageView { get; set; }

            public CaptureOutputDelegate(ScannerViewController parentVC)
            {
                parent = parentVC;
                width = (int)UIScreen.MainScreen.Bounds.Size.Width;
                height = (int)UIScreen.MainScreen.Bounds.Size.Height;

                imageView = new UIImageView();
                imageView.Frame = new CGRect(x: 100, y: 100, width: 200, height: 300);
                parent.View.AddSubview(imageView);
                parent.View.BringSubviewToFront(imageView);
            }
            public override void DidOutputSampleBuffer(AVCaptureOutput captureOutput, CMSampleBuffer sampleBuffer, AVCaptureConnection connection) {
                
                using (sampleBuffer) {
                    if (parent.state != eMainScreenState.CAMERA) return;

                    if (parent.state != eMainScreenState.CAMERA_DECODING) {
                        parent.state = eMainScreenState.CAMERA_DECODING;
                        //Console.WriteLine("Try to decode");

                        using (var imageBuffer = sampleBuffer.GetImageBuffer() as CVPixelBuffer) {
                            
                            //Get prepared 'framebuffer' for transfer into Barcode2DScanner.framework
                            imageBuffer.Lock(lockFlags: CVPixelBufferLock.None);
                            var baseAddres = (IntPtr)imageBuffer.BaseAddress;
                            CVPixelFormatType ft = imageBuffer.PixelFormatType;

                            int height = 0;
                            int width = 0;

                            byte[] frameBuffer;

                            switch (ft) {	// A different camera pixel formats should be taked into account
                                case CVPixelFormatType.CV420YpCbCr8BiPlanarVideoRange:

                                    width = (int)imageBuffer.GetBytesPerRowOfPlane(planeIndex: 0);
                                    height = (int)imageBuffer.GetHeightOfPlane(planeIndex: 0);
                                    frameBuffer = new byte[width * height];

                                    Marshal.Copy(baseAddres, frameBuffer, 0, height * width);
                                    break;

                                case CVPixelFormatType.CV422YpCbCr8:
                                    
                                    width = (int)imageBuffer.GetWidthOfPlane(planeIndex: 0);
                                    height = (int)imageBuffer.GetHeightOfPlane(planeIndex: 0);

                                    frameBuffer = new byte[width * height];
                                    byte[] bufferPixel = new byte[1];
                                    int j = 0;
                                    int bytePerPixel = (int)imageBuffer.GetBytesPerRowOfPlane(0) / width;
                                    for (int i = 0; i < height * width; i++)
                                    {
                                        Marshal.Copy(baseAddres + j, bufferPixel, 0, 1);
                                        j += bytePerPixel;
                                        //int newPixel = (bufferPixel[0] * bufferPixel[1] * bufferPixel[2]) / 3;
                                        int newPixel = bufferPixel[0];
                                        frameBuffer[i] = (byte)newPixel;
                                    }
                                    break;

                                default: frameBuffer = new byte[0];
                                    break;
                            }

                            imageBuffer.Unlock(unlockFlags: CVPixelBufferLock.None);
                            //Console.WriteLine("Array was copied");

                            //Call framework method to scan barcode from image
                            CoreFoundation.DispatchQueue.GetGlobalQueue(CoreFoundation.DispatchQueuePriority.Default).DispatchAsync(() =>
                            {
                            Barcode2DScanner scanner = new Barcode2DScanner();
                            scanner.RegisterCode("cDoLmB.HtIaPuHrDuEnNaA.EpKdNfG4L1K7E");
                            var handle = GCHandle.Alloc(value: frameBuffer, type: GCHandleType.Pinned);

                                try
                                {

									//! This is a main method of framework, which detect information from pacture.
									//! Please take into account a device orientation and barcode orientation in the you project.
									//! You should rotate buffer picture to the correct direction.
									//! Our example works only with portrait mode and vertical orientation of barcode.
                                    NSString result = (NSString)scanner.ScanGrayscaleImage(pp_image: handle.AddrOfPinnedObject(), width: width, height: height);

                                    Console.WriteLine("Result Length: {0}", result.Length);
                                    if (result.Length > 4)
                                    {
                                        parent.detected = true;

                                        CoreFoundation.DispatchQueue.MainQueue.DispatchAsync(() =>
                                        {
                                            parent.detected = false;
                                            parent.stopRunning();
                                        });

                                        UIApplication.SharedApplication.InvokeOnMainThread(delegate
                                        {
                                            UIAlertView alert = new UIAlertView()
                                            {
                                                Title = "",
                                                Message = result
                                            };
                                            alert.AddButton("OK");
                                            alert.Show();
                                            alert.Clicked += (sender, buttonArgs) => parent.startRunning();
                                        });
                                    }
                                    else
                                    {
                                        parent.detected = true;
                                        parent.state = eMainScreenState.CAMERA;
                                    }
                                }

                                finally {
                                    if (handle.IsAllocated)
                                    {
                                        handle.Free();
                                    }
                                }

                            }); //scanner
                        }   //using imageBuffer
                    }
                }   //using sampleBuffer
            }   //delegate method
        }   //delegate class
    }   //class ScannerViewController
}

